#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
DRAC RAID specific methods
"""

from oslo_log import log as logging
from oslo_utils import importutils
import sushy
import tenacity

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import metrics_utils
from ironic.common import states
from ironic.conductor import periodics
from ironic.conductor import utils as manager_utils
from ironic.conf import CONF
from ironic.drivers import base
from ironic.drivers.modules import deploy_utils
from ironic.drivers.modules.drac import utils as drac_utils
from ironic.drivers.modules.redfish import raid as redfish_raid
from ironic.drivers.modules.redfish import utils as redfish_utils

sushy_oem_idrac = importutils.try_import('sushy_oem_idrac')

LOG = logging.getLogger(__name__)

METRICS = metrics_utils.get_metrics_logger(__name__)


def _wait_till_realtime_ready(task):
    """Waits till real time operations are ready to be executed.

    Useful for RAID operations where almost all controllers support
    real time configuration, but controllers might not be ready for
    it by the time IPA starts executing steps. It can take minute or
    bit more to be ready for real time configuration.

    :param task: TaskManager object containing the node.
    :raises RedfishError: If can't find OEM extension or it fails to
        execute
    """
    # If running without IPA, check that system is ON, if not, turn it on
    disable_ramdisk = task.node.driver_internal_info.get(
        'cleaning_disable_ramdisk')
    power_state = task.driver.power.get_power_state(task)
    if disable_ramdisk and power_state == states.POWER_OFF:
        task.driver.power.set_power_state(task, states.POWER_ON)

    try:
        _retry_till_realtime_ready(task)
    except tenacity.RetryError:
        LOG.debug('Retries exceeded while waiting for real-time ready '
                  'for node %(node)s. Will proceed with out real-time '
                  'ready state', {'node': task.node.uuid})


@tenacity.retry(
    stop=(tenacity.stop_after_attempt(30)),
    wait=tenacity.wait_fixed(10),
    retry=tenacity.retry_if_result(lambda result: not result))
def _retry_till_realtime_ready(task):
    """Retries till real time operations are ready to be executed.

    :param task: TaskManager object containing the node.
    :raises RedfishError: If can't find OEM extension or it fails to
        execute
    :raises RetryError: If retries exceeded and still not ready for real-time
    """
    return _is_realtime_ready(task)


def _is_realtime_ready(task):
    """Gets is real time ready status

    Uses sushy-oem-idrac extension.

    :param task: TaskManager object containing the node.
    :returns: True, if real time operations are ready, otherwise False.
    :raises RedfishError: If can't find OEM extension or it fails to
        execute
    """
    return drac_utils.execute_oem_manager_method(
        task, 'get real-time ready status',
        lambda m: m.lifecycle_service.is_realtime_ready())


class DracRedfishRAID(redfish_raid.RedfishRAID):
    """iDRAC Redfish interface for RAID related actions.

    Includes iDRAC specific adjustments for RAID related actions.
    """

    @base.clean_step(priority=0, abortable=False, argsinfo={
        'create_root_volume': {
            'description': (
                'This specifies whether to create the root volume. '
                'Defaults to `True`.'
            ),
            'required': False
        },
        'create_nonroot_volumes': {
            'description': (
                'This specifies whether to create the non-root volumes. '
                'Defaults to `True`.'
            ),
            'required': False
        },
        'delete_existing': {
            'description': (
                'Setting this to `True` indicates to delete existing RAID '
                'configuration prior to creating the new configuration. '
                'Default value is `False`.'
            ),
            'required': False,
        }
    }, requires_ramdisk=False)
    def create_configuration(self, task, create_root_volume=True,
                             create_nonroot_volumes=True,
                             delete_existing=False):
        """Create RAID configuration on the node.

        This method creates the RAID configuration as read from
        node.target_raid_config.  This method
        by default will create all logical disks.

        :param task: TaskManager object containing the node.
        :param create_root_volume: Setting this to False indicates
            not to create root volume that is specified in the node's
            target_raid_config. Default value is True.
        :param create_nonroot_volumes: Setting this to False indicates
            not to create non-root volumes (all except the root volume) in
            the node's target_raid_config.  Default value is True.
        :param delete_existing: Setting this to True indicates to delete RAID
            configuration prior to creating the new configuration. Default is
            False.
        :returns: states.CLEANWAIT if RAID configuration is in progress
            asynchronously or None if it is complete.
        :raises: RedfishError if there is an error creating the configuration
        """
        _wait_till_realtime_ready(task)
        return super(DracRedfishRAID, self).create_configuration(
            task, create_root_volume, create_nonroot_volumes,
            delete_existing)

    @base.clean_step(priority=0, requires_ramdisk=False)
    @base.deploy_step(priority=0)
    def delete_configuration(self, task):
        """Delete RAID configuration on the node.

        :param task: TaskManager object containing the node.
        :returns: states.CLEANWAIT (cleaning) or states.DEPLOYWAIT (deployment)
            if deletion is in progress asynchronously or None if it is
            complete.
        """
        _wait_till_realtime_ready(task)
        return super(DracRedfishRAID, self).delete_configuration(task)

    def _validate_vendor(self, task):
        pass  # for now assume idrac-redfish is used with iDRAC BMC, thus pass

    def pre_create_configuration(self, task, logical_disks_to_create):
        """Perform required actions before creating config.

        Converts any physical disks of selected controllers to RAID mode
        if in non-RAID mode.

        :param task: a TaskManager instance containing the node to act on.
        :param logical_disks_to_create: list of logical disks to create.
        :returns: updated list of logical disks to create
        """
        system = redfish_utils.get_system(task.node)
        controller_to_disks = {}
        for logical_disk in logical_disks_to_create:
            storage, controller = DracRedfishRAID._get_storage_controller(
                system, logical_disk.get('controller'))
            controller_to_disks[controller] = []
            for drive in storage.drives:
                if drive.identity in logical_disk.get('physical_disks'):
                    controller_to_disks[controller].append(drive)

        converted = DracRedfishRAID._change_physical_disk_state(
            system,
            sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_RAID,
            controller_to_disks)

        if converted:
            # Recalculate sizes as disks size changes after conversion
            return DracRedfishRAID._get_revalidated_logical_disks(
                task.node, system, logical_disks_to_create)
        else:
            return logical_disks_to_create

    def post_delete_configuration(self, task, raid_configs, return_state=None):
        """Perform post delete_configuration action to commit the config.

        Clears foreign configuration for all RAID controllers.
        If no foreign configuration to clear, then checks if any controllers
        can be converted to RAID mode.

        :param task: a TaskManager instance containing the node to act on.
        :param raid_configs: a list of dictionaries containing the RAID
                             configuration operation details.
        :param return_state: state to return based on operation being invoked
        """

        system = redfish_utils.get_system(task.node)
        async_proc = DracRedfishRAID._clear_foreign_config(system, task)
        if async_proc:
            # Async processing with system rebooting in progress
            task.node.set_driver_internal_info(
                'raid_config_substep', 'clear_foreign_config')
            task.node.save()
            return deploy_utils.get_async_step_return_state(task.node)
        else:
            conv_state = DracRedfishRAID._convert_controller_to_raid_mode(
                task)
            if conv_state:
                return conv_state

        return return_state

    @staticmethod
    def _get_storage_controller(system, identity):
        """Finds storage and controller by identity

        :param system: Redfish system
        :param identity: identity of controller to find
        :returns: Storage and its controller
        """
        for storage in system.storage.get_members():
            if storage.identity == identity:
                controller = redfish_utils.get_first_controller(storage)
                if controller:
                    return storage, controller

        raise exception.IronicException(
            (_("Couldn't find storage by '%(identity)s'"),
             {'identity': identity}))

    @staticmethod
    def _change_physical_disk_state(system, mode, controller_to_disks=None):
        """Changes physical disk state and waits for it to complete

        :param system: Redfish system
        :param mode: sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_RAID or
            sushy_oem_idrac.PHYSICAL_DISK_STATE_MODE_NONRAID
        :controller_to_disks: dictionary of controllers and their
            drives. Optional. If not provided, then converting all
            eligible drives on system.
        :returns: True if any drive got converted, otherwise False
        """
        oem_sys = system.get_oem_extension('Dell')
        try:
            task_mons = oem_sys.change_physical_disk_state(
                mode, controller_to_disks)
        except AttributeError as ae:
            # For backported version where libraries could be too old
            LOG.warning('Failed to find method to convert drives to RAID '
                        'mode. Possibly because `sushy-oem-idrac` is too old. '
                        'Without newer `sushy-oem-idrac` RAID configuration '
                        'will fail if selected physical disks are in non-RAID '
                        'mode. To avoid that update `sushy-oem-idrac`. '
                        'Error: %(err)s', {'err': ae})
            return False

        for task_mon in task_mons:
            # All jobs should be real-time, because all RAID controllers
            # that offer physical disk mode conversion support real-time
            # task execution. Note that BOSS does not offer disk mode
            # conversion nor support real-time task execution.
            if task_mon.check_is_processing:
                task_mon.wait(CONF.drac.raid_job_timeout)

        return bool(task_mons)

    @staticmethod
    def _get_revalidated_logical_disks(
            node, system, logical_disks_to_create):
        """Revalidates calculated volume size after RAID mode conversion

        :param node: an Ironic node
        :param system: Redfish system
        :param logical_disks_to_create:
        :returns: Revalidated logical disk list. If no changes in size,
            same as input `logical_disks_to_create`
        """
        new_physical_disks, disk_to_controller =\
            redfish_raid.get_physical_disks(node)
        free_space_bytes = {}
        for disk in new_physical_disks:
            free_space_bytes[disk] = disk.capacity_bytes

        new_processed_volumes = []
        for logical_disk in logical_disks_to_create:
            selected_disks = [disk for disk in new_physical_disks
                              if disk.identity
                              in logical_disk['physical_disks']]

            spans_count = redfish_raid._calculate_spans(
                logical_disk['raid_level'], len(selected_disks))
            new_max_vol_size_bytes = redfish_raid._max_volume_size_bytes(
                logical_disk['raid_level'], selected_disks, free_space_bytes,
                spans_count=spans_count)
            if logical_disk['size_bytes'] > new_max_vol_size_bytes:
                logical_disk['size_bytes'] = new_max_vol_size_bytes
                LOG.info("Logical size does not match so calculating volume "
                         "properties for current logical_disk")
                redfish_raid._calculate_volume_props(
                    logical_disk, new_physical_disks, free_space_bytes,
                    disk_to_controller)
                new_processed_volumes.append(logical_disk)

        if new_processed_volumes:
            return new_processed_volumes

        return logical_disks_to_create

    @staticmethod
    def _clear_foreign_config(system, task):
        """Clears foreign config for given system

        :param system: Redfish system
        :param task: a TaskManager instance containing the node to act on
        :returns: True if system needs rebooting and async processing for
            tasks necessary, otherwise False
        """
        oem_sys = system.get_oem_extension('Dell')
        try:
            task_mons = oem_sys.clear_foreign_config()
        except AttributeError as ae:
            # For backported version where libraries could be too old
            LOG.warning('Failed to find method to clear foreign config. '
                        'Possibly because `sushy-oem-idrac` is too old. '
                        'Without newer `sushy-oem-idrac` no foreign '
                        'configuration will be cleared if there is any. '
                        'To avoid that update `sushy-oem-idrac`. '
                        'Error: %(err)s', {'err': ae})
            return False

        # Check if any of tasks requires reboot
        for task_mon in task_mons:
            oem_task = task_mon.get_task().get_oem_extension('Dell')
            if oem_task.job_type == sushy_oem_idrac.JOB_TYPE_RAID_CONF:
                # System rebooting, prepare ramdisk to boot back in IPA
                deploy_utils.set_async_step_flags(
                    task.node,
                    reboot=True,
                    skip_current_step=True,
                    polling=True)
                deploy_utils.prepare_agent_boot(task)
                # Reboot already done by non real time task
                task.upgrade_lock()
                task.node.set_driver_internal_info(
                    'raid_task_monitor_uris',
                    [tm.task_monitor_uri for tm in task_mons])
                task.node.save()
                return True

        # No task requiring reboot found, proceed with waiting for sync tasks
        for task_mon in task_mons:
            if task_mon.check_is_processing:
                task_mon.wait(CONF.drac.raid_job_timeout)
        return False

    @staticmethod
    def _convert_controller_to_raid_mode(task):
        """Convert eligible controllers to RAID mode if not already.

        :param task: a TaskManager instance containing the node to act on
        :returns: Return state if there are controllers to convert and
            and rebooting, otherwise None.
        """

        system = redfish_utils.get_system(task.node)
        task_mons = []
        warning_msg_templ = (
            'Possibly because `%(pkg)s` is too old. Without newer `%(pkg)s` '
            'PERC 9 and PERC 10 controllers that are not in RAID mode will '
            'not be used or have limited RAID support. To avoid that update '
            '`%(pkg)s`')
        for storage in system.storage.get_members():
            storage_controllers = None
            try:
                storage_controllers = storage.controllers
            except sushy.exceptions.MissingAttributeError:
                # Check if there storage_controllers to separate old iDRAC and
                # storage without controller
                if storage.storage_controllers:
                    LOG.warning('%(storage)s does not have controllers for '
                                'node %(node)s' + warning_msg_templ,
                                {'storage': storage.identity,
                                 'node': task.node.uuid,
                                 'pkg': 'iDRAC'})
                continue
            except AttributeError:
                LOG.warning('%(storage)s does not have controllers attribute. '
                            + warning_msg_templ, {'storage': storage.identity,
                                                  'pkg': 'sushy'})
                return None
            if storage_controllers:
                controller = storage.controllers.get_members()[0]
                try:
                    oem_controller = controller.get_oem_extension('Dell')
                except sushy.exceptions.ExtensionError as ee:
                    LOG.warning('Failed to find extension to convert '
                                'controller to RAID mode. '
                                + warning_msg_templ + '. Error: %(err)s',
                                {'err': ee, 'pkg': 'sushy-oem-idrac'})
                    return None
                task_mon = oem_controller.convert_to_raid()
                if task_mon:
                    task_mons.append(task_mon)

        if task_mons:
            deploy_utils.set_async_step_flags(
                task.node,
                reboot=True,
                skip_current_step=True,
                polling=True)

            task.upgrade_lock()
            task.node.set_driver_internal_info(
                'raid_task_monitor_uris',
                [tm.task_monitor_uri for tm in task_mons])
            task.node.save()
            return deploy_utils.reboot_to_finish_step(task)

    @METRICS.timer('DracRedfishRAID._query_raid_tasks_status')
    @periodics.node_periodic(
        purpose='checking async RAID tasks',
        spacing=CONF.drac.query_raid_config_job_status_interval,
        filters={'reserved': False, 'maintenance': False},
        predicate_extra_fields=['driver_internal_info'],
        predicate=lambda n: (
            n.driver_internal_info.get('raid_task_monitor_uris')
        ),
    )
    def _query_raid_tasks_status(self, task, manager, context):
        """Periodic task to check the progress of running RAID tasks"""
        self._check_raid_tasks_status(
            task, task.node.driver_internal_info.get('raid_task_monitor_uris'))

    def _check_raid_tasks_status(self, task, task_mon_uris):
        """Checks RAID tasks for completion

        If at least one of the jobs failed, then all step failed.
        If some tasks are still running, they are checked in next period.
        """
        node = task.node
        completed_task_mon_uris = []
        failed_msgs = []
        for task_mon_uri in task_mon_uris:
            task_mon = redfish_utils.get_task_monitor(node, task_mon_uri)
            if not task_mon.is_processing:
                raid_task = task_mon.get_task()
                completed_task_mon_uris.append(task_mon_uri)
                if not (raid_task.task_state == sushy.TASK_STATE_COMPLETED
                        and raid_task.task_status in
                        [sushy.HEALTH_OK, sushy.HEALTH_WARNING]):
                    messages = [m.message for m in raid_task.messages
                                if m.message is not None]
                    failed_msgs.append(
                        (_("Task %(task_mon_uri)s. "
                            "Message: '%(message)s'.")
                            % {'task_mon_uri': task_mon_uri,
                               'message': ', '.join(messages)}))

        task.upgrade_lock()
        if failed_msgs:
            error_msg = (_("Failed RAID configuration tasks: %(messages)s")
                         % {'messages': ', '.join(failed_msgs)})
            log_msg = ("RAID configuration task failed for node "
                       "%(node)s. %(error)s" % {'node': node.uuid,
                                                'error': error_msg})
            node.del_driver_internal_info('raid_task_monitor_uris')
            self._set_failed(task, log_msg, error_msg)
        else:
            running_task_mon_uris = [x for x in task_mon_uris
                                     if x not in completed_task_mon_uris]
            if running_task_mon_uris:
                node.set_driver_internal_info('raid_task_monitor_uris',
                                              running_task_mon_uris)
                # will check remaining jobs in the next period
            else:
                # all tasks completed and none of them failed
                node.del_driver_internal_info('raid_task_monitor_uris')
                substep = node.driver_internal_info.get(
                    'raid_config_substep')
                if substep == 'clear_foreign_config':
                    node.del_driver_internal_info('raid_config_substep')
                    node.save()
                    res = DracRedfishRAID._convert_controller_to_raid_mode(
                        task)
                    if res:  # New tasks submitted
                        return
                self._set_success(task)
        node.save()

    def _set_failed(self, task, log_msg, error_msg):
        if task.node.clean_step:
            manager_utils.cleaning_error_handler(task, log_msg, error_msg)
        else:
            manager_utils.deploying_error_handler(task, log_msg, error_msg)

    def _set_success(self, task):
        if task.node.clean_step:
            manager_utils.notify_conductor_resume_clean(task)
        else:
            manager_utils.notify_conductor_resume_deploy(task)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for the API /events methods.
"""

from http import client as http_client
from unittest import mock

from keystonemiddleware import auth_token
from oslo_config import cfg

from ironic.api.controllers import base as api_base
from ironic.api.controllers.v1 import event
from ironic.api.controllers.v1 import versions
from ironic.common import args
from ironic.common import exception
from ironic.tests import base as test_base
from ironic.tests.unit.api import base as test_api_base


def get_fake_port_event():
    return {'event': 'network.bind_port',
            'port_id': '11111111-aaaa-bbbb-cccc-555555555555',
            'mac_address': 'de:ad:ca:fe:ba:be',
            'status': 'ACTIVE',
            'device_id': '22222222-aaaa-bbbb-cccc-555555555555',
            'binding:host_id': '22222222-aaaa-bbbb-cccc-555555555555',
            'binding:vnic_type': 'baremetal'}


class TestEventValidator(test_base.TestCase):
    def setUp(self):
        super(TestEventValidator, self).setUp()
        self.v_event = event.NETWORK_EVENT_VALIDATOR
        self.v_events = args.schema(event.EVENTS_SCHEMA)

    def test_simple_event_type(self):
        self.v_events('body', {'events': [get_fake_port_event()]})

    def test_invalid_event_type(self):
        value = {'events': [{'event': 'invalid.event'}]}
        self.assertRaisesRegex(exception.Invalid,
                               "Schema error for body: "
                               "'invalid.event' is not one of",
                               self.v_events, 'body', value)

    def test_event_missing_madatory_field(self):
        value = {'invalid': 'invalid'}
        self.assertRaisesRegex(exception.Invalid,
                               "Schema error for event: "
                               "'event' is a required property",
                               self.v_event, 'event', value)

    def test_invalid_mac_network_port_event(self):
        value = {'event': 'network.bind_port',
                 'port_id': '11111111-aaaa-bbbb-cccc-555555555555',
                 'mac_address': 'INVALID_MAC_ADDRESS',
                 'status': 'ACTIVE',
                 'device_id': '22222222-aaaa-bbbb-cccc-555555555555',
                 'binding:host_id': '22222222-aaaa-bbbb-cccc-555555555555',
                 'binding:vnic_type': 'baremetal'
                 }
        self.assertRaisesRegex(exception.Invalid,
                               'Expected valid MAC address for mac_address: '
                               'INVALID_MAC_ADDRESS',
                               self.v_event, 'event', value)

    def test_missing_mandatory_fields_network_port_event(self):
        value = {'event': 'network.bind_port',
                 'device_id': '22222222-aaaa-bbbb-cccc-555555555555',
                 'binding:host_id': '22222222-aaaa-bbbb-cccc-555555555555',
                 'binding:vnic_type': 'baremetal'
                 }
        self.assertRaisesRegex(exception.Invalid,
                               "Schema error for event: "
                               "'port_id' is a required property",
                               self.v_event, 'event', value)


class TestPost(test_api_base.BaseApiTest):

    def setUp(self):
        super(TestPost, self).setUp()
        self.headers = {api_base.Version.string: str(
            versions.max_version_string())}

    def test_events(self):
        events_dict = {'events': [get_fake_port_event()]}
        response = self.post_json('/events', events_dict, headers=self.headers)
        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_multiple_events(self):
        events_dict = {'events': [get_fake_port_event(),
                                  get_fake_port_event(),
                                  get_fake_port_event()]}
        response = self.post_json('/events', events_dict, headers=self.headers)
        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_events_does_not_contain_event(self):
        events_dict = {'events': [{'INVALID': 'fake.event'}]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_events_invalid_event(self):
        events_dict = {'events': [{'event': 'invalid.event'}]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_unknown_event_property(self):
        events_dict = {'events': [{'event': 'network.unbind_port',
                                   'UNKNOWN': 'EVENT_PROPERTY'}]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_bind_port_events(self):
        events_dict = {'events': [get_fake_port_event()]}
        response = self.post_json('/events', events_dict, headers=self.headers)
        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_network_unbind_port_events(self):
        events_dict = {'events': [get_fake_port_event()]}
        events_dict['events'][0].update({'event': 'network.unbind_port'})
        response = self.post_json('/events', events_dict, headers=self.headers)
        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_network_delete_port_events(self):
        events_dict = {'events': [get_fake_port_event()]}
        events_dict['events'][0].update({'event': 'network.delete_port'})
        response = self.post_json('/events', events_dict, headers=self.headers)
        self.assertEqual(http_client.NO_CONTENT, response.status_int)

    def test_network_port_event_invalid_mac_address(self):
        port_evt = get_fake_port_event()
        port_evt.update({'mac_address': 'INVALID_MAC_ADDRESS'})
        events_dict = {'events': [port_evt]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_port_event_invalid_device_id(self):
        port_evt = get_fake_port_event()
        port_evt.update({'device_id': 'DEVICE_ID_SHOULD_BE_UUID'})
        events_dict = {'events': [port_evt]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_port_event_invalid_port_id(self):
        port_evt = get_fake_port_event()
        port_evt.update({'port_id': 'PORT_ID_SHOULD_BE_UUID'})
        events_dict = {'events': [port_evt]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_port_event_invalid_status(self):
        port_evt = get_fake_port_event()
        port_evt.update({'status': ['status', 'SHOULD', 'BE', 'TEXT']})
        events_dict = {'events': [port_evt]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_port_event_invalid_binding_vnic_type(self):
        port_evt = get_fake_port_event()
        port_evt.update({'binding:vnic_type': ['binding:vnic_type', 'SHOULD',
                                               'BE', 'TEXT']})
        events_dict = {'events': [port_evt]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_network_port_event_invalid_binding_host_id(self):
        port_evt = get_fake_port_event()
        port_evt.update({'binding:host_id': ['binding:host_id', 'IS',
                                             'NODE_UUID', 'IN', 'IRONIC']})
        events_dict = {'events': [port_evt]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=self.headers)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])

    def test_events_unsupported_api_version(self):
        headers = {api_base.Version.string: '1.50'}
        events_dict = {'events': [get_fake_port_event()]}
        response = self.post_json('/events', events_dict, expect_errors=True,
                                  headers=headers)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertTrue(response.json['error_message'])


@mock.patch.object(auth_token.AuthProtocol, 'process_request',
                   lambda *_: None)
class TestPostRBAC(TestPost):

    """Test class to execute the Event post tests with RBAC enforcement."""
    def setUp(self):
        super(TestPostRBAC, self).setUp()

        cfg.CONF.set_override('enforce_scope', True, group='oslo_policy')
        cfg.CONF.set_override('enforce_new_defaults', True,
                              group='oslo_policy')
        cfg.CONF.set_override('auth_strategy', 'keystone')
        # Headers required for this to pass in system scope restricted
        # authentication, as our default for api tests is noauth.
        self.headers = {
            api_base.Version.string: str(
                versions.max_version_string()),
            'X-Auth-Token': 'test-auth-token',
            'X-Roles': 'admin',
            'OpenStack-System-Scope': 'all',
        }



# -*- encoding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_context import context


CONF = cfg.CONF


class RequestContext(context.RequestContext):
    """Extends security contexts from the oslo.context library."""

    # NOTE(TheJulia): This is a flag used by oslo.context which allows us to
    # pass in a list of keys to preserve when calling from_dict() on the
    # RequestContext class.
    FROM_DICT_EXTRA_KEYS = ['auth_token_info']

    def __init__(self, is_public_api=False, auth_token_info=None, **kwargs):
        """Initialize the RequestContext

        :param is_public_api: Specifies whether the request should be processed
            without authentication.
        :param auth_token_info: Parameter to house auth token validation
            response data such as the user auth token's project id as opposed
            to the bearer token used. This allows for easy access to attributes
            for the end user when actions are taken on behalf of a user.
        :param kwargs: additional arguments passed to oslo.context.
        """
        super(RequestContext, self).__init__(**kwargs)
        self.is_public_api = is_public_api
        self.auth_token_info = auth_token_info

    def to_policy_values(self):
        policy_values = super(RequestContext, self).to_policy_values()
        policy_values.update({
            'project_name': self.project_name,
            'is_public_api': self.is_public_api,
        })
        if (CONF.rbac_service_role_elevated_access
                and CONF.rbac_service_project_name is not None):
            policy_values['config.service_project_name'] = CONF.rbac_service_project_name  # noqa
        return policy_values

    def ensure_thread_contain_context(self):
        """Ensure threading contains context

        For async/periodic tasks, the context of local thread is missing.
        Set it with request context and this is useful to log the request_id
        in log messages.

        """
        if context.get_current():
            return
        self.update_store()

    @classmethod
    def from_environ(cls, environ, **kwargs):
        """Load a context object from a request environment.

        If keyword arguments are provided then they override the values in the
        request environment, injecting the kwarg arguments used by ironic, as
        unknown values are filtered out from the final context object in
        the base oslo.context library.

        :param environ: The environment dictionary associated with a request.
        :type environ: dict
        """
        context = super().from_environ(environ)
        context.is_public_api = environ.get('is_public_api', False)
        context.auth_token_info = environ.get('keystone.token_info')
        return context

    def to_dict(self):
        """Return a dictionary of context attributes."""
        # The parent class in oslo.context provides the core standard
        # fields, but does not go beyond that. This preserves auth_token_info
        # for serialization and ultimately things like RPC transport.
        cdict = super().to_dict()
        cdict['auth_token_info'] = self.auth_token_info
        return cdict


def get_admin_context():
    """Create an administrator context."""

    context = RequestContext(auth_token=None,
                             project_id=None,
                             overwrite=False)
    return context



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests for the API /conductors/ methods.
"""

import datetime
from http import client as http_client
from unittest import mock

from oslo_config import cfg
from oslo_utils import timeutils
from oslo_utils import uuidutils

from ironic.api.controllers import base as api_base
from ironic.api.controllers import v1 as api_v1
from ironic.tests.unit.api import base as test_api_base
from ironic.tests.unit.objects import utils as obj_utils


class TestListConductors(test_api_base.BaseApiTest):

    def test_empty(self):
        data = self.get_json(
            '/conductors',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertEqual([], data['conductors'])

    def test_list(self):
        obj_utils.create_test_conductor(self.context, hostname='why care')
        obj_utils.create_test_conductor(self.context, hostname='why not')
        data = self.get_json(
            '/conductors',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertEqual(2, len(data['conductors']))
        for c in data['conductors']:
            self.assertIn('hostname', c)
            self.assertIn('conductor_group', c)
            self.assertIn('alive', c)
            self.assertNotIn('drivers', c)
        self.assertEqual(data['conductors'][0]['hostname'], 'why care')
        self.assertEqual(data['conductors'][1]['hostname'], 'why not')

    def test_list_with_detail(self):
        obj_utils.create_test_conductor(self.context, hostname='why care')
        obj_utils.create_test_conductor(self.context, hostname='why not')
        data = self.get_json(
            '/conductors?detail=true',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertEqual(2, len(data['conductors']))
        for c in data['conductors']:
            self.assertIn('hostname', c)
            self.assertIn('drivers', c)
            self.assertIn('conductor_group', c)
            self.assertIn('alive', c)
            self.assertIn('drivers', c)
        self.assertEqual(data['conductors'][0]['hostname'], 'why care')
        self.assertEqual(data['conductors'][1]['hostname'], 'why not')

    def test_list_with_invalid_api(self):
        response = self.get_json(
            '/conductors', headers={api_base.Version.string: '1.48'},
            expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)

    def test_get_one(self):
        obj_utils.create_test_conductor(self.context, hostname='rocky.rocks')
        data = self.get_json(
            '/conductors/rocky.rocks',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertIn('hostname', data)
        self.assertIn('drivers', data)
        self.assertIn('conductor_group', data)
        self.assertIn('alive', data)
        self.assertIn('drivers', data)
        self.assertEqual(data['hostname'], 'rocky.rocks')
        self.assertTrue(data['alive'])

    def test_get_one_with_port_and_v6(self):
        obj_utils.create_test_conductor(self.context, hostname='[::1]:8090')
        data = self.get_json(
            '/conductors/[::1]:8090',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertIn('hostname', data)
        self.assertIn('drivers', data)
        self.assertIn('conductor_group', data)
        self.assertIn('alive', data)
        self.assertIn('drivers', data)
        self.assertEqual(data['hostname'], '[::1]:8090')
        self.assertTrue(data['alive'])

    @mock.patch.object(timeutils, 'utcnow', autospec=True)
    def test_get_one_conductor_offline_old_heartbeat(self, mock_utcnow):
        self.config(heartbeat_timeout=10, group='conductor')

        _time = datetime.datetime(2000, 1, 1, 0, 0)
        mock_utcnow.return_value = _time

        obj_utils.create_test_conductor(self.context, hostname='rocky.rocks')

        mock_utcnow.return_value = _time + datetime.timedelta(seconds=30)

        data = self.get_json(
            '/conductors/rocky.rocks',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertIn('hostname', data)
        self.assertIn('drivers', data)
        self.assertIn('conductor_group', data)
        self.assertIn('alive', data)
        self.assertIn('drivers', data)
        self.assertEqual(data['hostname'], 'rocky.rocks')
        self.assertFalse(data['alive'])

    def test_get_one_conductor_offline_unregistered(self):
        conductor = obj_utils.create_test_conductor(
            self.context, hostname='rocky.rocks')
        conductor.unregister()

        data = self.get_json(
            '/conductors/rocky.rocks',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertFalse(data['alive'])

    def test_get_one_with_invalid_api(self):
        response = self.get_json(
            '/conductors/rocky.rocks',
            headers={api_base.Version.string: '1.48'},
            expect_errors=True)
        self.assertEqual(http_client.NOT_FOUND, response.status_int)

    def test_get_one_custom_fields(self):
        obj_utils.create_test_conductor(self.context, hostname='rocky.rocks')
        fields = 'hostname,alive'
        data = self.get_json(
            '/conductors/rocky.rocks?fields=%s' % fields,
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertCountEqual(['hostname', 'alive', 'links'], data)

    def test_get_collection_custom_fields(self):
        obj_utils.create_test_conductor(self.context, hostname='rocky.rocks')
        obj_utils.create_test_conductor(self.context, hostname='stein.rocks')
        fields = 'hostname,alive'

        data = self.get_json(
            '/conductors?fields=%s' % fields,
            headers={api_base.Version.string: str(api_v1.max_version())})

        self.assertEqual(2, len(data['conductors']))
        for c in data['conductors']:
            self.assertCountEqual(['hostname', 'alive', 'links'], c)

    def test_get_custom_fields_invalid_fields(self):
        obj_utils.create_test_conductor(self.context, hostname='rocky.rocks')
        fields = 'hostname,spongebob'
        response = self.get_json(
            '/conductors/rocky.rocks?fields=%s' % fields,
            headers={api_base.Version.string: str(api_v1.max_version())},
            expect_errors=True)
        self.assertEqual(http_client.BAD_REQUEST, response.status_int)
        self.assertEqual('application/json', response.content_type)
        self.assertIn('spongebob', response.json['error_message'])

    def _test_links(self, public_url=None):
        cfg.CONF.set_override('public_endpoint', public_url, 'api')
        obj_utils.create_test_conductor(self.context, hostname='rocky.rocks')
        headers = {api_base.Version.string: str(api_v1.max_version())}
        data = self.get_json(
            '/conductors/rocky.rocks',
            headers=headers)
        self.assertIn('links', data)
        self.assertEqual(2, len(data['links']))
        self.assertIn('rocky.rocks', data['links'][0]['href'])
        for link in data['links']:
            bookmark = link['rel'] == 'bookmark'
            self.assertTrue(self.validate_link(link['href'], bookmark=bookmark,
                                               headers=headers))

        if public_url is not None:
            expected = [{'href': '%s/v1/conductors/rocky.rocks' % public_url,
                         'rel': 'self'},
                        {'href': '%s/conductors/rocky.rocks' % public_url,
                         'rel': 'bookmark'}]
            for i in expected:
                self.assertIn(i, data['links'])

    def test_links(self):
        self._test_links()

    def test_links_public_url(self):
        self._test_links(public_url='http://foo')

    def test_collection_links(self):
        conductors = []
        for id in range(5):
            hostname = uuidutils.generate_uuid()
            conductor = obj_utils.create_test_conductor(self.context,
                                                        hostname=hostname)
            conductors.append(conductor.hostname)
        data = self.get_json(
            '/conductors/?limit=3',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertEqual(3, len(data['conductors']))

        next_marker = data['conductors'][-1]['hostname']
        self.assertIn('/conductors', data['next'])
        self.assertIn('limit=3', data['next'])
        self.assertIn(f'marker={next_marker}', data['next'])

    def test_collection_links_default_limit(self):
        cfg.CONF.set_override('max_limit', 3, 'api')
        conductors = []
        for id in range(5):
            hostname = uuidutils.generate_uuid()
            conductor = obj_utils.create_test_conductor(self.context,
                                                        hostname=hostname)
            conductors.append(conductor.hostname)
        data = self.get_json(
            '/conductors',
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertEqual(3, len(data['conductors']))

        next_marker = data['conductors'][-1]['hostname']
        self.assertIn('/conductors', data['next'])
        self.assertIn(f'marker={next_marker}', data['next'])

    def test_collection_links_custom_fields(self):
        cfg.CONF.set_override('max_limit', 3, 'api')
        conductors = []
        fields = 'hostname,alive'
        for id in range(5):
            hostname = uuidutils.generate_uuid()
            conductor = obj_utils.create_test_conductor(self.context,
                                                        hostname=hostname)
            conductors.append(conductor.hostname)
        data = self.get_json(
            '/conductors?fields=%s' % fields,
            headers={api_base.Version.string: str(api_v1.max_version())})
        self.assertEqual(3, len(data['conductors']))

        next_marker = data['conductors'][-1]['hostname']
        self.assertIn('/conductors', data['next'])
        self.assertIn(f'marker={next_marker}', data['next'])
        self.assertIn(f'fields={fields}', data['next'])

    def test_sort_key(self):
        conductors = []
        for id in range(5):
            hostname = uuidutils.generate_uuid()
            conductor = obj_utils.create_test_conductor(self.context,
                                                        hostname=hostname)
            conductors.append(conductor.hostname)
        data = self.get_json(
            '/conductors?sort_key=hostname',
            headers={api_base.Version.string: str(api_v1.max_version())})
        hosts = [n['hostname'] for n in data['conductors']]
        self.assertEqual(sorted(conductors), hosts)

    def test_sort_key_invalid(self):
        invalid_keys_list = ['alive', 'drivers']
        headers = {api_base.Version.string: str(api_v1.max_version())}
        for invalid_key in invalid_keys_list:
            response = self.get_json('/conductors?sort_key=%s' % invalid_key,
                                     headers=headers,
                                     expect_errors=True)
            self.assertEqual(http_client.BAD_REQUEST, response.status_int)
            self.assertEqual('application/json', response.content_type)
            self.assertIn(invalid_key, response.json['error_message'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_utils import units

from ironic.common import exception
from ironic.conductor import task_manager
from ironic.conf import CONF
from ironic.drivers.modules.inspector.hooks import root_device as \
    root_device_hook
from ironic.tests.unit.db import base as db_base
from ironic.tests.unit.objects import utils as obj_utils


class RootDeviceTestCase(db_base.DbTestCase):
    def setUp(self):
        super().setUp()
        CONF.set_override('enabled_inspect_interfaces',
                          ['agent', 'no-inspect'])
        self.node = obj_utils.create_test_node(self.context,
                                               inspect_interface='agent')
        self.inventory = {'disks': [
            {'name': '/dev/sda', 'model': 'Model 1', 'size': 1000 * units.Gi,
             'serial': '1111'},
            {'name': '/dev/sdb', 'model': 'Model 2', 'size': 10 * units.Gi,
             'serial': '2222'},
            {'name': '/dev/sdc', 'model': 'Model 1', 'size': 20 * units.Gi,
             'serial': '3333'},
            {'name': '/dev/sdd', 'model': 'Model 3', 'size': 0,
             'serial': '4444'}]}
        self.plugin_data = {}

    def test_no_hints(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            self.assertNotIn('root_disk', self.plugin_data)
            self.assertEqual(self.plugin_data['local_gb'], 0)
            self.assertEqual(task.node.properties.get('local_gb'), '0')

    def test_root_device_skip_list(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'serial': '1111'}
            task.node.properties['skip_block_devices'] = [{'size': 1000}]

            self.assertRaisesRegex(exception.HardwareInspectionFailure,
                                   'No disks satisfied root device hints for '
                                   'node %s' % self.node.id,
                                   root_device_hook.RootDeviceHook().__call__,
                                   task, self.inventory, self.plugin_data)
            self.assertNotIn('root_disk', self.plugin_data)
            self.assertNotIn('local_gb', self.plugin_data)
            # The default value of the `local_gb` property is left unchanged
            self.assertEqual(task.node.properties.get('local_gb'), '10')

    def test_first_match_on_skip_list_use_second(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'model': 'Model 1'}
            task.node.properties['skip_block_devices'] = [{'size': 1000}]
            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            task.node.refresh()

            expected_root_device = self.inventory['disks'][2].copy()
            self.assertEqual(self.plugin_data['root_disk'],
                             expected_root_device)

            expected_local_gb = (expected_root_device['size'] // units.Gi) - 1
            self.assertEqual(self.plugin_data['local_gb'],
                             expected_local_gb)
            self.assertEqual(task.node.properties.get('local_gb'),
                             str(expected_local_gb))

    def test_one_matches(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'serial': '1111'}

            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            task.node.refresh()

            self.assertEqual(self.plugin_data['root_disk'],
                             self.inventory['disks'][0])
            self.assertEqual(self.plugin_data['local_gb'], 999)
            self.assertEqual(task.node.properties.get('local_gb'), '999')

    def test_local_gb_without_spacing(self):
        CONF.set_override('disk_partitioning_spacing', False, 'inspector')
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'serial': '1111'}
            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            task.node.refresh()

            self.assertEqual(self.plugin_data['root_disk'],
                             self.inventory['disks'][0])
            self.assertEqual(self.plugin_data['local_gb'], 1000)
            self.assertEqual(task.node.properties.get('local_gb'), '1000')

    def test_zero_size(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'name': '/dev/sdd'}
            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            task.node.refresh()
            self.assertEqual(self.plugin_data['root_disk'],
                             self.inventory['disks'][3])
            self.assertEqual(self.plugin_data['local_gb'], 0)
            self.assertEqual(task.node.properties.get('local_gb'), '0')

    def test_all_match(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'size': 20,
                                                   'model': 'Model 1'}
            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            task.node.refresh()
            self.assertEqual(self.plugin_data['root_disk'],
                             self.inventory['disks'][2])
            self.assertEqual(self.plugin_data['local_gb'], 19)
            self.assertEqual(task.node.properties.get('local_gb'), '19')

    def test_incorrect_hint(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'size': 20,
                                                   'model': 'Model 42'}
            self.assertRaisesRegex(exception.HardwareInspectionFailure,
                                   'No disks satisfied root device hints for '
                                   'node %s' % task.node.uuid,
                                   root_device_hook.RootDeviceHook().__call__,
                                   task, self.inventory, self.plugin_data)
            self.assertNotIn('root_disk', self.plugin_data)
            self.assertNotIn('local_gb', self.plugin_data)
            # The default value of the `local_gb` property is unchanged
            self.assertEqual(task.node.properties.get('local_gb'), '10')

    def test_size_string(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            task.node.properties['root_device'] = {'size': '20'}
            root_device_hook.RootDeviceHook().__call__(task,
                                                       self.inventory,
                                                       self.plugin_data)
            task.node.refresh()
            self.assertEqual(self.plugin_data['root_disk'],
                             self.inventory['disks'][2])
            self.assertEqual(self.plugin_data['local_gb'], 19)
            self.assertEqual(task.node.properties.get('local_gb'), '19')

    def test_size_invalid(self):
        with task_manager.acquire(self.context, self.node.id) as task:
            for bad_size in ('foo', None, {}):
                task.node.properties['root_device'] = {'size': bad_size}
                self.assertRaisesRegex(
                    exception.HardwareInspectionFailure,
                    'No disks could be found using root device hints',
                    root_device_hook.RootDeviceHook().__call__,
                    task, self.inventory, self.plugin_data)

                self.assertNotIn('root_disk', self.plugin_data)
                self.assertNotIn('local_gb', self.plugin_data)
                # The default value of the `local_gb` property is left
                # unchanged
                self.assertEqual(task.node.properties.get('local_gb'), '10')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

from ironic.common import exception
from ironic.common.i18n import _
from ironic.common import utils
from ironic.drivers.modules.inspector.hooks import base

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class RootDeviceHook(base.InspectionHook):
    """Smarter root disk selection using Ironic root device hints."""

    def _get_skip_list_for_node(self, node, block_devices):
        skip_list_hints = node.properties.get("skip_block_devices", [])
        if not skip_list_hints:
            return
        skip_list = set()

        for hint in skip_list_hints:
            skipped_devs = utils.find_devices_by_hints(block_devices, hint)
            excluded_devs = {dev['name'] for dev in skipped_devs}
            skipped_devices = excluded_devs.difference(skip_list)
            skip_list = skip_list.union(excluded_devs)

            if skipped_devices:
                LOG.warning("Using hint %(hint)s skipping devices: %(devs)s",
                            {'hint': hint, 'devs': ','.join(skipped_devices)})
        return skip_list

    def _process_root_device_hints(self, node, inventory, plugin_data):
        """Detect root disk from root device hints and IPA inventory."""

        hints = node.properties.get('root_device')
        if not hints:
            LOG.debug('Root device hints are not provided for node %s',
                      node.uuid)
            return

        skip_list = self._get_skip_list_for_node(node, inventory['disks'])
        if skip_list:
            inventory_disks = [d for d in inventory['disks']
                               if d['name'] not in skip_list]
        else:
            inventory_disks = inventory['disks']

        try:
            root_device = utils.match_root_device_hints(inventory_disks,
                                                        hints)
        except (TypeError, ValueError) as e:
            raise exception.HardwareInspectionFailure(
                _('No disks could be found using root device hints %(hints)s '
                  'for node %(node)s because they failed to validate. '
                  'Error: %(error)s') % {'hints': hints, 'node': node.uuid,
                                         'error': e})
        if not root_device:
            raise exception.HardwareInspectionFailure(_(
                'No disks satisfied root device hints for node %s') %
                node.uuid)

        LOG.debug('Disk %(disk)s of size %(size)s satisfies root device '
                  'hints. Node: %s', {'disk': root_device.get('name'),
                                      'size': root_device['size'],
                                      'node': node.uuid})
        plugin_data['root_disk'] = root_device

    def __call__(self, task, inventory, plugin_data):
        """Process root disk information."""
        self._process_root_device_hints(task.node, inventory, plugin_data)

        root_disk = plugin_data.get('root_disk')
        if root_disk:
            local_gb = root_disk['size'] // units.Gi
            if not local_gb:
                LOG.warning('The requested root disk is too small (smaller '
                            'than 1 GiB) or its size cannot be detected. '
                            'Root disk: %s, Node: %s', root_disk,
                            task.node.uuid)
            else:
                if CONF.inspector.disk_partitioning_spacing:
                    local_gb -= 1
                LOG.info('Root disk %(disk)s, local_gb %(local_gb)s GiB, '
                         'Node: %(node)s', {'disk': root_disk,
                                            'local_gb': local_gb,
                                            'node': task.node.uuid})
        else:
            local_gb = 0
            LOG.info('No root device found for node %s. Assuming node is '
                     'diskless.', task.node.uuid)

        plugin_data['local_gb'] = local_gb
        task.node.set_property('local_gb', str(local_gb))
        task.node.save()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Port Physical Network Hook"""

import ipaddress

from oslo_config import cfg
from oslo_log import log as logging

from ironic.drivers.modules.inspector.hooks import base
from ironic import objects

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class PhysicalNetworkHook(base.InspectionHook):
    """Hook to set the port's physical_network field.

    Set the ironic port's physical_network field based on a CIDR to physical
    network mapping in the configuration.
    """

    dependencies = ['validate-interfaces']

    def get_physical_network(self, interface):
        """Return a physical network to apply to an ironic port.

        :param interface: The interface from the inventory.
        :returns: The physical network to set, or None.
        """

        def get_interface_ips(iface):
            ips = []
            for addr_version in ['ipv4_address', 'ipv6_address']:
                try:
                    ips.append(ipaddress.ip_address(iface.get(addr_version)))
                except ValueError:
                    pass
            return ips

        # Convert list config to a dictionary with ip_networks as keys
        cidr_map = {
            ipaddress.ip_network(x.rsplit(':', 1)[0]): x.rsplit(':', 1)[1]
            for x in CONF.inspector.physical_network_cidr_map}
        ips = get_interface_ips(interface)
        for ip in ips:
            try:
                return [cidr_map[cidr] for cidr in cidr_map if ip in cidr][0]
            except IndexError:
                # This IP address is not present in the CIDR map
                pass
        # No mapping found for any of the IP addresses
        return None

    def __call__(self, task, inventory, plugin_data):
        """Process inspection data and patch the port's physical network."""

        node_ports = objects.Port.list_by_node_id(task.context, task.node.id)
        ports_dict = {p.address: p for p in node_ports}

        for interface in inventory['interfaces']:
            if interface['name'] not in plugin_data['all_interfaces']:
                continue

            mac_address = interface['mac_address']
            port = ports_dict.get(mac_address)
            if not port:
                LOG.debug("Skipping physical network processing for interface "
                          "%s on node %s - matching port not found in Ironic.",
                          mac_address, task.node.uuid)
                continue

            # Determine the physical network for this port, using the interface
            # IPs and CIDR map configuration.
            phys_network = self.get_physical_network(interface)
            if phys_network is None:
                LOG.debug("Skipping physical network processing for interface "
                          "%s on node %s - no physical network mapping.",
                          mac_address,
                          task.node.uuid)
                continue

            if getattr(port, 'physical_network', '') != phys_network:
                port.physical_network = phys_network
                port.save()
                LOG.debug('Updated physical_network of port %s to %s',
                          port.uuid, port.physical_network)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ironic.conf import types
from ironic.tests.base import TestCase


class ExplicitAbsolutePath(TestCase):
    def test_explicit_absolute_path(self):
        """Verifies the Opt subclass used to validate absolute paths."""
        good_paths = [
            '/etc/passwd',  # Valid
            '/usr/bin/python',  # Valid
            '/home/user/file.txt',  # Valid - dot in filename allowed
            '/var/lib/ironic/.secretdir',  # Valid - hidden directory allowed
            '/var/lib/ironic/oslo.config',  # Valid - dots in filename allowed
            '/tmp/',  # Valid
            '/',  # Valid (root directory)
            '/.hidden_root_file',  # Valid
            '/path/including/a/numb3r',  # Valid
            '/a/path/with/a/trailing/slash/'  # Valid
        ]
        bad_paths = [
            'relative/path',  # Invalid - no leading slash
            './file.txt',  # Invalid - relative path
            '../file.txt',  # Invalid - relative path
            'file.txt',  # Invalid - no leading slash
            '',  # Invalid - empty string
            '/var/lib/ironic/../../../etc/passwd',  # Invalid - path traversal
            '/etc/../etc/passwd',  # Invalid - path traversal
            '/home/user/./config',  # Invalid - contains current dir reference
            '/home/user/../user/config',  # Invalid - path traversal
            '/../etc/passwd',  # Invalid - path traversal at beginning
            '/.',  # Invalid - just current directory
            '/..'  # Invalid - just parent directory
        ]

        eap = types.ExplicitAbsolutePath()

        def _trypath(tpath):
            try:
                eap(tpath)
            except ValueError:
                return False
            else:
                return True

        for path in good_paths:
            self.assertTrue(_trypath(path),
                            msg=f"Improperly disallowed path: {path}")

        for path in bad_paths:
            self.assertFalse(_trypath(path),
                             msg=f"Improperly allowed path: {path}")



# coding=utf-8
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import netutils
from oslo_utils import strutils
from oslo_utils import uuidutils
from oslo_utils import versionutils
from oslo_versionedobjects import base as object_base

from ironic.common import exception
from ironic.common import utils
from ironic.db import api as dbapi
from ironic.objects import base
from ironic.objects import fields as object_fields
from ironic.objects import notification


@base.IronicObjectRegistry.register
class Port(base.IronicObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    # Version 1.1: Add get() and get_by_id() and get_by_address() and
    #              make get_by_uuid() only work with a uuid
    # Version 1.2: Add create() and destroy()
    # Version 1.3: Add list()
    # Version 1.4: Add list_by_node_id()
    # Version 1.5: Add list_by_portgroup_id() and new fields
    #              local_link_connection, portgroup_id and pxe_enabled
    # Version 1.6: Add internal_info field
    # Version 1.7: Add physical_network field
    # Version 1.8: Migrate/copy extra['vif_port_id'] to
    #              internal_info['tenant_vif_port_id'] (not an explicit db
    #              change)
    # Version 1.9: Add support for Smart NIC port
    # Version 1.10: Add name field
    # Version 1.11: Add node_uuid field
    VERSION = '1.11'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(),
        'uuid': object_fields.UUIDField(nullable=True),
        'node_id': object_fields.IntegerField(nullable=True),
        'node_uuid': object_fields.UUIDField(nullable=True),
        'address': object_fields.MACAddressField(nullable=True),
        'extra': object_fields.FlexibleDictField(nullable=True),
        'local_link_connection': object_fields.FlexibleDictField(
            nullable=True),
        'portgroup_id': object_fields.IntegerField(nullable=True),
        'pxe_enabled': object_fields.BooleanField(),
        'internal_info': object_fields.FlexibleDictField(nullable=True),
        'physical_network': object_fields.StringField(nullable=True),
        'is_smartnic': object_fields.BooleanField(nullable=True,
                                                  default=False),
        'name': object_fields.StringField(nullable=True),
    }

    def _convert_name_field(self, target_version,
                            remove_unavailable_fields=True):
        name_is_set = self.obj_attr_is_set('name')
        if target_version >= (1, 10):
            # Target version supports name. Set it to its default
            # value if it is not set.
            if not name_is_set:
                self.name = None
        elif name_is_set:
            # Target version does not support name, and it is set.
            if remove_unavailable_fields:
                # (De)serialising: remove unavailable fields.
                delattr(self, 'name')
            elif self.name is not None:
                # DB: set unavailable fields to their default.
                self.name = None

    def _convert_to_version(self, target_version,
                            remove_unavailable_fields=True):
        """Convert to the target version.

        Convert the object to the target version. The target version may be
        the same, older, or newer than the version of the object. This is
        used for DB interactions as well as for serialization/deserialization.

        Version 1.7: physical_network field was added. Its default value is
            None. For versions prior to this, it should be set to None (or
            removed).

        Version 1.8: if extra['vif_port_id'] is specified (non-null) and
            internal_info['tenant_vif_port_id'] is not specified, copy the
            .extra value to internal_info. There is nothing to do here when
            downgrading to an older version.

        Version 1.9: remove is_smartnic field for unsupported versions if
            remove_unavailable_fields is True.

        Version 1.10: remove name field for unsupported versions if
            remove_unavailable_fields is True.

        :param target_version: the desired version of the object
        :param remove_unavailable_fields: True to remove fields that are
            unavailable in the target version; set this to True when
            (de)serializing. False to set the unavailable fields to appropriate
            values; set this to False for DB interactions.
        """
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version >= (1, 8):
            if self.obj_attr_is_set('extra'):
                vif = self.extra.get('vif_port_id')
                if vif:
                    internal_info = (self.internal_info
                                     if self.obj_attr_is_set('internal_info')
                                     else {})
                    if 'tenant_vif_port_id' not in internal_info:
                        internal_info['tenant_vif_port_id'] = vif
                        self.internal_info = internal_info

        # Convert the physical_network field.
        physnet_is_set = self.obj_attr_is_set('physical_network')
        if target_version >= (1, 7):
            # Target version supports physical_network. Set it to its default
            # value if it is not set.
            if not physnet_is_set:
                self.physical_network = None
        elif physnet_is_set:
            # Target version does not support physical_network, and it is set.
            if remove_unavailable_fields:
                # (De)serialising: remove unavailable fields.
                delattr(self, 'physical_network')
            elif self.physical_network is not None:
                # DB: set unavailable fields to their default.
                self.physical_network = None

        # Convert is_smartnic field.
        is_smartnic_set = self.obj_attr_is_set('is_smartnic')
        if target_version >= (1, 9):
            # Target version supports is_smartnic. Set it to its default
            # value if it is not set.
            if not is_smartnic_set:
                self.is_smartnic = False

        # handle is_smartnic field in older version
        elif is_smartnic_set:
            # Target version does not support is_smartnic, and it is set.
            if remove_unavailable_fields:
                # (De)serialising: remove unavailable fields.
                delattr(self, 'is_smartnic')
            elif self.is_smartnic is not False:
                # DB: set unavailable fields to their default.
                self.is_smartnic = False

        # Convert the name field.
        self._convert_name_field(target_version, remove_unavailable_fields)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get(cls, context, port_id):
        """Find a port.

        Find a port based on its id or uuid or name or MAC address and return
        a Port object.

        :param context: Security context
        :param port_id: the id *or* uuid *or* name *or* MAC address of a port.
        :returns: a :class:`Port` object.
        :raises: InvalidIdentity

        """
        if strutils.is_int_like(port_id):
            return cls.get_by_id(context, port_id)
        elif uuidutils.is_uuid_like(port_id):
            return cls.get_by_uuid(context, port_id)
        elif netutils.is_valid_mac(port_id):
            return cls.get_by_address(context, port_id)
        elif utils.is_valid_logical_name(port_id):
            return cls.get_by_name(context, port_id)
        else:
            raise exception.InvalidIdentity(identity=port_id)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_id(cls, context, port_id):
        """Find a port based on its integer ID and return a Port object.

        :param cls: the :class:`Port`
        :param context: Security context
        :param port_id: the ID of a port.
        :returns: a :class:`Port` object.
        :raises: PortNotFound

        """
        db_port = cls.dbapi.get_port_by_id(port_id)
        port = cls._from_db_object(context, cls(), db_port)
        return port

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a port based on UUID and return a :class:`Port` object.

        :param cls: the :class:`Port`
        :param context: Security context
        :param uuid: the UUID of a port.
        :returns: a :class:`Port` object.
        :raises: PortNotFound

        """
        db_port = cls.dbapi.get_port_by_uuid(uuid)
        port = cls._from_db_object(context, cls(), db_port)
        return port

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_address(cls, context, address, owner=None, project=None):
        """Find a port based on address and return a :class:`Port` object.

        :param cls: the :class:`Port`
        :param context: Security context
        :param address: the address of a port.
        :param owner: DEPRECATED a node owner to match against
        :param project: a node owner or lessee to match against
        :returns: a :class:`Port` object.
        :raises: PortNotFound

        """
        if owner and not project:
            project = owner
        db_port = cls.dbapi.get_port_by_address(address, project=project)
        port = cls._from_db_object(context, cls(), db_port)
        return port

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_name(cls, context, name):
        """Find a port based on name and return a :class:`Port` object.

        :param cls: the :class:`Port`
        :param context: Security context
        :param name: the name of a port.
        :returns: a :class:`Port` object.
        :raises: PortNotFound

        """
        db_port = cls.dbapi.get_port_by_name(name)
        port = cls._from_db_object(context, cls(), db_port)
        return port

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def list(cls, context, limit=None, marker=None,
             sort_key=None, sort_dir=None, owner=None, project=None):
        """Return a list of Port objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param owner: DEPRECATED a node owner to match against
        :param project: a node owner or lessee to match against
        :returns: a list of :class:`Port` object.
        :raises: InvalidParameterValue

        """
        if owner and not project:
            project = owner
        db_ports = cls.dbapi.get_port_list(limit=limit,
                                           marker=marker,
                                           sort_key=sort_key,
                                           sort_dir=sort_dir,
                                           project=project)
        return cls._from_db_object_list(context, db_ports)

    @classmethod
    def list_by_node_shards(cls, context, shards, limit=None, marker=None,
                            sort_key=None, sort_dir=None, project=None):
        """Return a list of Port objects associated with nodes in shards

        :param context: Security context.
        :param shards: a list of shards
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param project: a node owner or lessee to match against
        :returns: a list of :class:`Port` object.

        """
        db_ports = cls.dbapi.get_ports_by_shards(shards, limit=limit,
                                                 marker=marker,
                                                 sort_key=sort_key,
                                                 sort_dir=sort_dir)
        return cls._from_db_object_list(context, db_ports)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def list_by_node_id(cls, context, node_id, limit=None, marker=None,
                        sort_key=None, sort_dir=None, owner=None,
                        project=None):
        """Return a list of Port objects associated with a given node ID.

        :param context: Security context.
        :param node_id: the ID of the node.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param owner: DEPRECATED a node owner to match against
        :param project: a node owner or lessee to match against
        :returns: a list of :class:`Port` object.

        """
        if owner and not project:
            project = owner
        db_ports = cls.dbapi.get_ports_by_node_id(node_id, limit=limit,
                                                  marker=marker,
                                                  sort_key=sort_key,
                                                  sort_dir=sort_dir,
                                                  project=project)
        return cls._from_db_object_list(context, db_ports)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def list_by_portgroup_id(cls, context, portgroup_id, limit=None,
                             marker=None, sort_key=None, sort_dir=None,
                             owner=None, project=None):
        """Return a list of Port objects associated with a given portgroup ID.

        :param context: Security context.
        :param portgroup_id: the ID of the portgroup.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param owner: DEPRECATED a node owner to match against
        :param project: a node owner or lessee to match against
        :returns: a list of :class:`Port` object.

        """
        if owner and not project:
            project = owner
        db_ports = cls.dbapi.get_ports_by_portgroup_id(portgroup_id,
                                                       limit=limit,
                                                       marker=marker,
                                                       sort_key=sort_key,
                                                       sort_dir=sort_dir,
                                                       project=project)
        return cls._from_db_object_list(context, db_ports)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def create(self, context=None):
        """Create a Port record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Port(context)
        :raises: MACAlreadyExists if 'address' column is not unique
        :raises: PortAlreadyExists if 'uuid' column is not unique

        """
        values = self.do_version_changes_for_db()
        db_port = self.dbapi.create_port(values)
        # NOTE(hjensas): To avoid lazy load issue (DetachedInstanceError) in
        # sqlalchemy, get new port the port from the DB to ensure the node_uuid
        # via association_proxy relationship is loaded.
        db_port = self.dbapi.get_port_by_id(db_port['id'])
        self._from_db_object(self._context, self, db_port)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def destroy(self, context=None):
        """Delete the Port from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Port(context)
        :raises: PortNotFound

        """
        self.dbapi.destroy_port(self.uuid)
        self.obj_reset_changes()

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def save(self, context=None):
        """Save updates to this Port.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Port(context)
        :raises: PortNotFound
        :raises: MACAlreadyExists if 'address' column is not unique

        """
        updates = self.do_version_changes_for_db()
        updated_port = self.dbapi.update_port(self.uuid, updates)
        self._from_db_object(self._context, self, updated_port)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def refresh(self, context=None):
        """Loads updates for this Port.

        Loads a port with the same uuid from the database and
        checks for updated attributes. Updates are applied from
        the loaded port column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Port(context)
        :raises: PortNotFound

        """
        current = self.get_by_uuid(self._context, uuid=self.uuid)
        self.obj_refresh(current)
        self.obj_reset_changes()

    @classmethod
    def supports_physical_network(cls):
        """Return whether the physical_network field is supported.

        :returns: Whether the physical_network field is supported
        :raises: ovo_exception.IncompatibleObjectVersion
        """
        return cls.supports_version((1, 7))

    @classmethod
    def supports_is_smartnic(cls):
        """Return whether is_smartnic field is supported.

        :returns: Whether is_smartnic field is supported
        :raises: ovo_exception.IncompatibleObjectVersion
        """
        return cls.supports_version((1, 9))

    def set_local_link_connection(self, key, value):
        """Set a `local_link_connection` value.

        Setting a `local_link_connection` dict value via this method ensures
        that this field will be flagged for saving.

        :param key: Key of item to set
        :param value: Value of item to set
        """
        self.local_link_connection[key] = value
        self._changed_fields.add('local_link_connection')


@base.IronicObjectRegistry.register
class PortCRUDNotification(notification.NotificationBase):
    """Notification emitted when ironic creates, updates or deletes a port."""
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'payload': object_fields.ObjectField('PortCRUDPayload')
    }


@base.IronicObjectRegistry.register
class PortCRUDPayload(notification.NotificationPayloadBase):
    # Version 1.0: Initial version
    # Version 1.1: Add "portgroup_uuid" field
    # Version 1.2: Add "physical_network" field
    # Version 1.3: Add "is_smartnic" field
    # Version 1.4: Add "name" field
    VERSION = '1.4'

    SCHEMA = {
        'address': ('port', 'address'),
        'extra': ('port', 'extra'),
        'local_link_connection': ('port', 'local_link_connection'),
        'pxe_enabled': ('port', 'pxe_enabled'),
        'physical_network': ('port', 'physical_network'),
        'created_at': ('port', 'created_at'),
        'updated_at': ('port', 'updated_at'),
        'uuid': ('port', 'uuid'),
        'is_smartnic': ('port', 'is_smartnic'),
        'name': ('port', 'name'),
    }

    fields = {
        'address': object_fields.MACAddressField(nullable=True),
        'extra': object_fields.FlexibleDictField(nullable=True),
        'local_link_connection': object_fields.FlexibleDictField(
            nullable=True),
        'pxe_enabled': object_fields.BooleanField(nullable=True),
        'node_uuid': object_fields.UUIDField(),
        'portgroup_uuid': object_fields.UUIDField(nullable=True),
        'physical_network': object_fields.StringField(nullable=True),
        'created_at': object_fields.DateTimeField(nullable=True),
        'updated_at': object_fields.DateTimeField(nullable=True),
        'uuid': object_fields.UUIDField(),
        'is_smartnic': object_fields.BooleanField(nullable=True,
                                                  default=False),
        'name': object_fields.StringField(nullable=True),
    }

    def __init__(self, port, node_uuid, portgroup_uuid):
        super(PortCRUDPayload, self).__init__(node_uuid=node_uuid,
                                              portgroup_uuid=portgroup_uuid)
        self.populate_schema(port=port)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import strutils
from oslo_utils import uuidutils
from oslo_versionedobjects import base as object_base

from ironic.common import exception
from ironic.db import api as dbapi
from ironic.objects import base
from ironic.objects import fields as object_fields


@base.IronicObjectRegistry.register
class NodeHistory(base.IronicObject, object_base.VersionedObjectDictCompat):
    # Version 1.0: Initial version
    VERSION = '1.0'

    dbapi = dbapi.get_instance()

    fields = {
        'id': object_fields.IntegerField(),
        'uuid': object_fields.UUIDField(nullable=True),
        'conductor': object_fields.StringField(nullable=True),
        'event': object_fields.StringField(nullable=True),
        'user': object_fields.StringField(nullable=True),
        'node_id': object_fields.IntegerField(nullable=True),
        'event_type': object_fields.StringField(nullable=True),
        'severity': object_fields.StringField(nullable=True),
    }

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get(cls, context, history_ident):
        """Get a history based on its id or uuid.

        :param history_ident: The id or uuid of a history.
        :param context: Security context
        :returns: A :class:`NodeHistory` object.
        :raises: InvalidIdentity

        """
        if strutils.is_int_like(history_ident):
            return cls.get_by_id(context, history_ident)
        elif uuidutils.is_uuid_like(history_ident):
            return cls.get_by_uuid(context, history_ident)
        else:
            raise exception.InvalidIdentity(identity=history_ident)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_id(cls, context, history_id):
        """Get a NodeHistory object by its integer ID.

        :param cls: the :class:`NodeHistory`
        :param context: Security context
        :param history_id: The ID of a history.
        :returns: A :class:`NodeHistory` object.
        :raises: NodeHistoryNotFound

        """
        db_history = cls.dbapi.get_node_history_by_id(history_id)
        history = cls._from_db_object(context, cls(), db_history)
        return history

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def get_by_uuid(cls, context, uuid):
        """Get a NodeHistory object by its UUID.

        :param cls: the :class:`NodeHistory`
        :param context: Security context
        :param uuid: The UUID of a NodeHistory.
        :returns: A :class:`NodeHistory` object.
        :raises: NodeHistoryNotFound

        """
        db_history = cls.dbapi.get_node_history_by_uuid(uuid)
        history = cls._from_db_object(context, cls(), db_history)
        return history

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def list(cls, context, limit=None, marker=None, sort_key=None,
             sort_dir=None):
        """Return a list of NodeHistory objects.

        :param cls: the :class:`NodeHistory`
        :param context: Security context.
        :param limit: Maximum number of resources to return in a single result.
        :param marker: Pagination marker for large data sets.
        :param sort_key: Column to sort results by.
        :param sort_dir: Direction to sort. "asc" or "desc".
        :returns: A list of :class:`NodeHistory` object.
        :raises: InvalidParameterValue

        """
        db_histories = cls.dbapi.get_node_history_list(limit=limit,
                                                       marker=marker,
                                                       sort_key=sort_key,
                                                       sort_dir=sort_dir)
        return cls._from_db_object_list(context, db_histories)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable_classmethod
    @classmethod
    def list_by_node_id(cls, context, node_id, limit=None, marker=None,
                        sort_key=None, sort_dir=None):
        """Return a list of NodeHistory objects belongs to a given node ID.

        :param cls: the :class:`NodeHistory`
        :param context: Security context.
        :param node_id: The ID of the node.
        :param limit: Maximum number of resources to return in a single result.
        :param marker: Pagination marker for large data sets.
        :param sort_key: Column to sort results by.
        :param sort_dir: Direction to sort. "asc" or "desc".
        :returns: A list of :class:`NodeHistory` object.
        :raises: InvalidParameterValue

        """
        db_histories = cls.dbapi.get_node_history_by_node_id(
            node_id, limit=limit, marker=marker, sort_key=sort_key,
            sort_dir=sort_dir)
        return cls._from_db_object_list(context, db_histories)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def create(self, context=None):
        """Create a NodeHistory record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: NodeHistory(context)
        """
        values = self.do_version_changes_for_db()
        db_history = self.dbapi.create_node_history(values)
        self._from_db_object(self._context, self, db_history)

    # NOTE(xek): We don't want to enable RPC on this call just yet. Remotable
    # methods can be used in the future to replace current explicit RPC calls.
    # Implications of calling new remote procedures should be thought through.
    # @object_base.remotable
    def destroy(self, context=None):
        """Delete the NodeHistory from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: NodeHistory(context)
        :raises: NodeHistoryNotFound
        """
        self.dbapi.destroy_node_history_by_uuid(self.uuid)
        self.obj_reset_changes()
